"""
Just shortens some lines to make code look better.
"""

FONT_ATTRIBUTES = ['font_name', 'font_size', 'font_bold', 'font_bold']
ANY_QUESTIONS = 'In case you have any questions, please ask them at https://github.com/Rednax-Code/pygame_ui/discussions'
ISSUE_REPORT = '[Note] if the above statement seems incorrect, please report this issue at https://github.com/Rednax-Code/pygame_ui/issues. Thank you :)\n'
MUST_BE = 'Must be one of the following:\n'
MORE_INFO = 'Check the documentation for more information about '
ALPHABET = ['a','b','c','d','e','f','g','h','i','j','k','l','m','n','o','p','q','r','s','t','u','v','w','x','y','z']
ANCHORS = {'top': -1, 'bottom': 1, 'left': -1, 'right': 1, 'center': 0}


"""
blabla
"""

import os
import sys

import pygame.font as pygfont
import pygame.rect as pygrect
import pygame.draw as pygdraw
import pygame.display as pygdisplay
import pygame.surface as pygsurface
import pygame._sdl2 as pygsdl2

from pygame_ui.constants import *


sys.path.append(os.getcwd())

pygfont.init()


class UI_Element:
	is_visible = True
	is_hoverable = False
	is_clickable = False
	position = [0,0]
	position_anchor = "top left"
	position_relative = False
	position_units = ["px","px"] # or ["w%", "h%"]
	size = [0,0]
	size_units = ["px", "px"] # or ["w%", "h%"]
	auto_size = False
	background_color = None
	use_sdl2 = False

	def __init__(self, initial_data, kwargs):
		if initial_data[0]['use_sdl2']:
			renderer = initial_data[0].pop('renderer')
		for dictionary in initial_data:
			for key in dictionary:
				setattr(self, key, dictionary[key])
		self.parent = kwargs.pop('parent')

		for key in kwargs:
			setattr(self, key, kwargs[key])
		
		self.position_initial = list(self.position)
		self.size_initial = list(self.size)
		self.change_position(self.position, self.position_units)
		self.change_size(self.size, self.size_units)

		if self.use_sdl2:
			if self.background_color != None:
				if len(self.background_color) < 4:
					bg_c = self.background_color + [255]
				else:
					bg_c = self.background_color
				image = pygsurface.Surface(self.size).convert_alpha()
				image.fill(bg_c)
				self.texture = pygsdl2.Texture.from_surface(renderer, image)

	def change_position(self, new_position, units=['px','px']):
		"""
		You can figure this one out yourself. If you can't then... *sigh*
		"""
		new_position = list(new_position)

		# Apply relative positioning
		position_zero = [0,0]
		if self.position_relative:
			percentage_target = self.parent.size
			position_zero = list(self.parent.position)
		else:
			percentage_target = pygdisplay.get_surface().get_size()

		# Apply anchor point
		if self.position_anchor == 'center':
			anchor = [0,0]
		else:
			anchor = [ANCHORS[self.position_anchor.split(' ')[i]] for i in [1,0]]
		
		# Apply percentages
		for i in [0, 1]:
			if units[i] == 'w%':
				new_position[i] = percentage_target[0]*new_position[i]/100
			elif units[i] == 'h%':
				new_position[i] = percentage_target[1]*new_position[i]/100
		
		self.position = [new_position[i] - (anchor[i]/2+.5)*self.size[i] + position_zero[i] for i in [0,1]]
		self.rectangle = pygrect.Rect(self.position, self.size)

	def change_size(self, new_size, units=['px','px']):
		"""
		You can figure this one out yourself. If you can't then... *sigh*
		"""
		new_size = list(new_size)
		
		# Apply relative positioning
		new_position = list(self.position)
		if self.position_relative:
			new_position = [self.position[i] - self.parent.position[i] for i in [0,1]]

		# Apply anchor point
		if self.position_anchor == 'center':
			anchor = [0,0]
		else:
			anchor = [ANCHORS[self.position_anchor.split(' ')[i]] for i in [1,0]]
		new_position = [new_position[i] + (anchor[i]/2+.5)*self.size[i] for i in [0,1]]

		# Apply percentages 
		percentage_target = pygdisplay.get_surface().get_size()
		for i in [0, 1]:
			if units[i] == 'w%':
				new_size[i] = percentage_target[0]*new_size[i]/100
			elif units[i] == 'h%':
				new_size[i] = percentage_target[1]*new_size[i]/100
		
		self.size = new_size

		# Calculate new position
		self.change_position(new_position)
	
	def draw_bg(self, pygame_window):
		draw_surface = pygsurface.Surface(pygdisplay.get_surface().get_size()).convert_alpha()
		draw_surface.fill((0,0,0,0))
		
		pygdraw.rect(draw_surface, self.background_color, self.rectangle)
		
		if len(self.background_color) < 4:
			draw_surface.set_alpha(255)

		pygame_window.blit(draw_surface, [0,0])
	
	def draw_bg_sdl2(self, renderer):
		# This works but i don't like it, it's not pretty and could probably be faster
		renderer.blit(self.texture, self.rectangle)
		# Something like this below would make more sense, but i couldn't get the alpha channel to work :(
		#renderer.draw_color = bg_c
		#renderer.alpha = bg_c[3]
		#renderer.fill_rect(self.rectangle)


class frame(UI_Element):
	"""
	The UI element that works as a sort of bundle or group.

	Yes, frame-ception is a thing.
	"""
	
	contents = {}

	def __init__(self, *initial_data, **kwargs):
		self.elements = {}
		if initial_data[0]['use_sdl2']:
			renderer = initial_data[0]['renderer']
		super().__init__(initial_data, kwargs)
		for name, data in self.contents.items():
			element_type = data.pop('type')
			data['use_sdl2'] = bool(self.use_sdl2)
			if self.use_sdl2:
				data['renderer'] = renderer
			self.elements[name] = globals()[element_type](data, parent=self)
	
	def get_text_input_elements(self):
		text_input_elements = []
		for name, element in self.elements.items():
			if element.is_visible:
				if isinstance(element, frame):
					text_input_elements.extend(element.get_text_input_elements())
				elif isinstance(element, text_input):
					text_input_elements.append(element)
		return text_input_elements

	def get_interactive_elements(self):
		interactives = []
		for name, element in self.elements.items():
			if element.is_visible:
				if isinstance(element, frame) and not isinstance(element, button):
					interactives.extend(element.get_interactive_elements())
				elif element.is_hoverable or element.is_clickable:
					interactives.append(element)
		return interactives
	
	def draw(self, pygame_window):
		for i in self.elements.values():
			if i.is_visible:
				if i.background_color != None:
					i.draw_bg(pygame_window)
				i.draw(pygame_window)
	
	def draw_sdl2(self, renderer):
		for i in self.elements.values():
			if i.is_visible:
				if i.background_color != None:
					i.draw_bg_sdl2(renderer)
				i.draw_sdl2(renderer)


class label(UI_Element):
	"""
	The UI element for text.

	Anything with letters will use this class.
	"""
	
	text = "Empty Label"
	text_color = (255,255,255)
	text_aa = True
	font_name = 'Arial'
	font_size = 10
	font_bold = False
	font_italic = False

	def __init__(self, *initial_data, **kwargs):
		super().__init__(initial_data, kwargs)
		self.font = pygfont.SysFont(self.font_name, self.font_size, self.font_bold, self.font_italic)
		if self.auto_size:
			if isinstance(self, label):
				self.change_size(self.font.size(self.text))
				self.rectangle = pygrect.Rect(self.position, self.size)
	
	def change_font(self, **kwargs):
		"""
		Used to change attributes of the font.

		The keyword arguments must exist in the following list:
		>>> ['font_name', 'font_size', 'font_bold', 'font_bold']
		"""

		for key in kwargs:
			if key in FONT_ATTRIBUTES:
				setattr(self, key, kwargs[key])
			else:
				raise KeyError(key+' is not a font attribute. '+MUST_BE+FONT_ATTRIBUTES+'\n'+ISSUE_REPORT)
		self.font = pygfont.SysFont(self.font_name, self.font_size, self.font_bold, self.font_italic)

	def draw(self, pygame_window):
		text_surface = self.font.render(self.text, self.text_aa, self.text_color, self.background_color)
		pygame_window.blit(text_surface, self.position)
	
	def draw_sdl2(self, renderer):
		text_surface = self.font.render(self.text, self.text_aa, self.text_color, self.background_color)
		if not 0 in text_surface.get_size():
			texture = pygsdl2.video.Texture.from_surface(renderer, text_surface)
			renderer.blit(texture, text_surface.get_rect().move(self.position))


class text_input(label):
	"""
	The UI element for an input field.

	My quote-bucket is empty :(
	"""
	is_clickable = True
	click_start = False
	click_end = False
	click_held = False
	typing_start_on_click = True
	typing_end_on_enter = True
	typing = False
	text = "Your input here"
	caret = False
	caret_timer = 0

	def draw(self, pygame_window):
		text_to_render = self.text
		if self.caret:
			text_to_render += '|'
		text_surface = self.font.render(text_to_render, self.text_aa, self.text_color, self.background_color)
		pygame_window.blit(text_surface, self.position)
	
	def draw_sdl2(self, renderer):
		text_to_render = self.text
		if self.caret:
			text_to_render += '|'
		text_surface = self.font.render(text_to_render, self.text_aa, self.text_color, self.background_color)
		if not 0 in text_surface.get_size():
			texture = pygsdl2.video.Texture.from_surface(renderer, text_surface)
			renderer.blit(texture, text_surface.get_rect().move(self.position))

class button(frame):
	"""
	The UI element for a button.

	What does this button do?
	"""
	is_clickable = True
	is_hoverable = True
	click_start = False
	click_end = False
	click_held = False
	hover_start = False
	hover_end = False
	hover_held = False


class switch(UI_Element):
	"""
	The UI element for a switch.

	Switcharoo!
	"""
	is_clickable = True
	is_hoverable = True
	click_start = False
	click_end = False
	click_held = False
	hover_start = False
	hover_end = False
	hover_held = False
	state = False
	preset = None

	def __init__(self, *initial_data, **kwargs):
		super().__init__(initial_data, kwargs)

	def draw(self, pygame_window):
		if self.preset == 'simple':
			size = [self.size[0]/2-5, self.size[1]-10]
			position = [self.position[0]+5+size[0]*int(self.state), self.position[1]+5]
			pygdraw.rect(pygame_window, (200,200,200), pygrect.Rect(position, size))

	def draw_sdl2(self, renderer):
		if self.preset == 'simple':
			size = [self.size[0]/2-5, self.size[1]-10]
			position = [self.position[0]+5+size[0]*int(self.state), self.position[1]+5]
			renderer.draw_color = (200,200,200,255)
			renderer.fill_rect(pygrect.Rect(position, size))

class slider(UI_Element):
	"""
	The UI element that slides.

	Sliiide to the left, sliiide to the right, criss cross!
	"""
	is_clickable = True
	is_hoverable = True
	click_start = False
	click_end = False
	click_held = False
	hover_start = False
	hover_end = False
	hover_held = False
	value_min = 0
	value_max = 1
	value = 0
	preset = None

	def __init__(self, *initial_data, **kwargs):
		super().__init__(initial_data, kwargs)

	def set_value_from_pos(self, position):
		if self.preset == 'simple':
			self.value/(self.value_max-self.value_min)
			x = (self.value_max - self.value_min) * (position[0]-self.size[1]/2 - self.position[0])/(self.size[0]-self.size[1])
			self.value = max(min(self.value_max, x), self.value_min)

	def draw(self, pygame_window):
		if self.preset == 'simple':
			size = [self.size[1]-10, self.size[1]-10]
			position = [self.position[0]+5+(self.size[0]-self.size[1])*self.value/(self.value_max-self.value_min), self.position[1]+5]
			pygdraw.rect(pygame_window, (200,200,200), pygrect.Rect(position, size))
	
	def draw_sdl2(self, renderer):
		if self.preset == 'simple':
			size = [self.size[1]-10, self.size[1]-10]
			position = [self.position[0]+5+(self.size[0]-self.size[1])*self.value/(self.value_max-self.value_min), self.position[1]+5]
			renderer.draw_color = (200,200,200,255)
			renderer.fill_rect(pygrect.Rect(position, size))


class dropdown(UI_Element):
	"""
	Unfinished
	"""
	is_clickable = True

	def __init__(self, *initial_data, **kwargs):
		super().__init__(initial_data, kwargs)

	def draw(self, pygame_window):
		pass


"""
Pygame UI
=========

A python library for quickly building user interfaces with pygame in a JSON.

Make sure you read the ``Getting started`` section of the documentation at https://pypi.org/project/pygame-json-ui/ or https://github.com/Rednax-Code/pygame_ui.
"""

from pygame_ui.base import *
from pygame_ui.elements import *

# release, year, month, session
__version__ = '1.0.1.0'

print('pygame-json-ui ' + __version__)

def test():
	try:
		import pygame_ui.base
		import pygame_ui.constants
		import pygame_ui.elements
		print('Successfully installed! enjoy :)\nversion: '+__version__+'\n'+ANY_QUESTIONS)
		del pygame_ui.base, pygame_ui.constants, pygame_ui.elements
	except:
		print(ISSUE_REPORT)

# Base64
# VGhpcyBjb2RlIHdhcyB3cml0dGVuIGJ5IFJlZG5heEdhbWluZyBvbiBnaXRodWI=


from pygame_ui.constants import *
import pygame_ui.elements
import pygame.locals as pyglocal
import pygame.mouse as pygmouse
import pygame.scrap as pygscrap
import pygame.key as pygkey
from time import time


class Graphical_UI:
	"""
	The main GUI handler.

	Is automatically called on ``pygame_ui.init()``

	Calling this manually is deprecated.
	"""

	elements = {}
	interactive_elements = []
	text_input_elements = []
	held_buttons = []
	backspace_hold_timer = 0
	old_time = time()
	new_time = time()

	def __init__(self, objects:dict, use_sdl2=False, renderer=None):
		self.use_sdl2 = use_sdl2
		for name, data in objects.items():
			element_type = data.pop('type')
			data['use_sdl2'] = use_sdl2
			if renderer != None:
				data['renderer'] = renderer
			self.elements[name] = getattr(pygame_ui.elements, element_type)(data, parent=None)

	def get_frame(self, frame_path:str=''):
		"""
		Returns the frame of given path.
		"""
		
		frame = self
		for i in frame_path.split('->'):
			if i in frame.elements.keys():
				frame = frame.elements[i]
			else:
				raise KeyError('No frame with name \''+i+'\' was found. '+MORE_INFO+'frame paths')
		return frame

	def get_element(self, name:str, frame_path:str=''):
		"""
		Returns the element with specified name at given path.
		"""

		frame = self.get_frame(frame_path)
		if name in frame.elements.keys():
			return frame.elements[name]
		else:
			raise KeyError('No element with name \''+name+'\' was found. '+MORE_INFO+'frame paths')

	def add_element(self, name:str, element:pygame_ui.elements.UI_Element, frame_path:str=''):
		"""
		Adds the specified element with it's name to the GUI or given frame.
		"""
		
		frame = self.get_frame(frame_path)
		frame.elements[name] = element

	def remove_element(self, name:str, frame_path:str=''):
		"""
		Removes the element specified by name (and path) from it's parent.
		"""

		frame = self.get_frame(frame_path)
		if name in frame.elements.keys():
			frame.elements.pop(name)
		else:
			raise KeyError('No element with name \''+frame_path+'\' was found. '+MORE_INFO+'frame paths')

	def get_interactive_elements(self):
		interactives = []
		for name, element in self.elements.items():
			if element.is_visible:
				if isinstance(element, pygame_ui.frame):
					interactives.extend(element.get_interactive_elements())
				elif element.is_hoverable or element.is_clickable:
					interactives.append(element)
		self.interactive_elements = interactives
	
	def video_resize(self, frame=None):
		"""
		Makes sure all elements with percentage positions or sizes will get updated to the new window.
		Usually called on event pygame.VIDEORESIZE.
		(Doesn't work with moving elements yet)
		"""
		if frame == None:
			target = self
		else:
			target = frame
		for name, element in target.elements.items():
			element.change_position(element.position_initial, element.position_units)
			if element.auto_size:
				if isinstance(element, pygame_ui.label):
					element.change_size(element.font.size(element.text))
			else:
				element.change_size(element.size_initial, element.size_units)
			if isinstance(element, pygame_ui.frame):
				self.video_resize(element)
		return 1

	def event_handler(self, event):
		"""
		This handles all interactive elements.

		Must be called in the following context:
		>>> for event in pygame.event.get():
			Interface.event_handler(event)
		"""

		# Gather held buttons
		if event.type == pyglocal.KEYDOWN:
			self.held_buttons.append(event.key)
		elif event.type == pyglocal.KEYUP:
			self.held_buttons.remove(event.key)

		# Handle mouse button presses
		self.get_interactive_elements()

		element = None
		lmb, rmb, mmb = pygmouse.get_pressed(3)
		mpos = pygmouse.get_pos()
		
		for i in self.interactive_elements:
			mouse_in_boundry = i.rectangle.collidepoint(mpos)

			# Add exception to hoverable elements to fix hover_end
			exception = False
			if i.is_hoverable:
				if i.hover_held:
					exception = True

			if mouse_in_boundry or exception:
				if event.type in [pyglocal.MOUSEBUTTONDOWN, pyglocal.MOUSEBUTTONUP] and i.is_clickable:
					element = i
				elif event.type in [pyglocal.MOUSEMOTION, pyglocal.MOUSEWHEEL] and i.is_hoverable:
					element = i
			elif event.type in [pyglocal.MOUSEBUTTONUP, pyglocal.MOUSEMOTION, pyglocal.MOUSEWHEEL] and i.click_held:
				element = i
			
			if event.type == pyglocal.MOUSEBUTTONUP and lmb == 0:
				i.click_end = True
				i.click_held = False

		if element != None:
			if event.type in [pyglocal.MOUSEMOTION, pyglocal.MOUSEWHEEL]:
				if element.is_hoverable:
					if element.rectangle.collidepoint(mpos) and not element.hover_held:
						element.hover_start = True
						element.hover_held = True
					elif not element.rectangle.collidepoint(mpos) and element.hover_held:
						element.hover_end = True
						element.hover_held = False
					
				if isinstance(element, pygame_ui.slider) and element.click_held:
					element.set_value_from_pos(mpos)

			elif event.type == pyglocal.MOUSEBUTTONDOWN and lmb == 1:
				element.click_start = True
				element.click_held = True
				if isinstance(element, pygame_ui.switch):
					element.state = not element.state
				elif isinstance(element, pygame_ui.text_input):
					if element.typing_start_on_click:
						if element.typing == False:
							element.text = ''
						element.typing = True
						element.caret = True


		# Handle keyboard inputs
		self.text_input_elements = []
		for name, element in self.elements.items():
			if element.is_visible:
				if isinstance(element, pygame_ui.frame):
					self.text_input_elements.extend(element.get_text_input_elements())

		if event.type == pyglocal.KEYDOWN:
			key = pygkey.name(event.key)
			for i in self.text_input_elements:
				if i.typing:
					i.caret = True
					i.caret_timer = 0
					if key in ALPHABET:
						if event.mod & pyglocal.KMOD_CTRL:
							if key == 'x':
								pygscrap.put(pyglocal.SCRAP_TEXT, bytes(i.text, 'utf-8'))
								i.text = ''
							elif key == 'c':
								pygscrap.put(pyglocal.SCRAP_TEXT, bytes(i.text, 'utf-8'))
							elif key == 'v':
								i.text += str(pygscrap.get(pyglocal.SCRAP_TEXT), 'utf-8')
						elif event.mod & pyglocal.KMOD_SHIFT:
							i.text += key.upper()
						else:
							i.text += key
					elif key == 'space':
						i.text += ' '
					elif key == 'backspace':
						i.text = i.text[:-1]
					elif key == 'return' and i.typing_end_on_enter:
						i.caret = False
						i.typing = False
		return 1
	
	def draw(self, pygame_window):
		"""
		Main draw function which will draw all visible things layered like painter-style.
		
		Must be called in the following context:
		>>> Interface.draw(pygame_window)
		>>> pygame.display.flip()

		Note: when using sdl2 instead of passing pygame_window, pass the renderer.
		"""

		# Calculate delta-time
		self.new_time = time()
		dtime = (self.new_time-self.old_time)*1000
		self.old_time = time()

		# Update all timers for the vertical line of text input elements
		for i in self.text_input_elements:
			if i.typing:
				i.caret_timer += dtime
				if i.caret_timer >= 520:
					i.caret = not i.caret
					i.caret_timer = 0

		if self.use_sdl2:
			for i in self.elements.values():
				if i.is_visible:
					if i.background_color != None:
						i.draw_bg_sdl2(pygame_window)
					i.draw_sdl2(pygame_window)
		else:
			for i in self.elements.values():
				if i.is_visible:
					if i.background_color != None:
						i.draw_bg(pygame_window)
					i.draw(pygame_window)
		
		# Reset all temporary attribute values
		for i in self.interactive_elements:
			if i.is_clickable:
				i.click_start = False
				i.click_end = False
			if i.is_hoverable:
				i.hover_start = False
				i.hover_end = False
				
		return 1


def init(path_to_json:str='Interface.json', use_sdl2=False, renderer=None):
	"""
	This loads in the ``Interface.json`` file and created a GUI with it
	"""

	import json
	import os
	
	try:
		file = open(os.path.abspath(path_to_json))
	except FileNotFoundError:
		print(ISSUE_REPORT)
	UI = json.load(file)
	file.close()

	interface = Graphical_UI(UI, use_sdl2, renderer)

	# clearing up namespace
	del json, os, pygame_ui.elements

	# Initialize the pygame module for access to clipboard
	pygscrap.init()

	return interface